#-----------------------------------------------------
# D-BUS bindings for Ruby
# (C) Copyright 2004 Leon Breedt
# (C) Copyright 2004 Provenco Group Ltd
#
# Licensed under the same terms as the D-BUS library
#-----------------------------------------------------

require 'dbus.so'
require 'dbus/version'
require 'dbus/binding'

module DBus

  # Represents a connection to a DBus daemon
  class Bus
    TYPE_SESSION = BUS_SESSION
    TYPE_SYSTEM = BUS_SYSTEM
    TYPE_ACTIVATION = BUS_ACTIVATION

    def initialize(bus_type=TYPE_SESSION, glib_mainloop=true)
      @connection = DBus::Binding::bus_get(bus_type)
      @connection.add_filter(method(:signal_dispatcher))
      @rules = {}
      @connection.setup_with_g_main if glib_mainloop
    end

    def get_connection
      @connection
    end

    def get_service(name="org.freedesktop.Broadcast")
      RemoteService.new(self, name)
    end

    # Add a handler +handler_proc+ for the match rule resulting from
    # passing the parameters to Bus#build_signal_rule. Multiple handlers
    # can be installed for the same signal.
    def add_signal_receiver(handler_proc, signal_name=nil, interface=nil, service=nil, path=nil)
      rule = build_signal_rule(signal_name, interface, service, path)
      @rules[rule] ||= []
      @rules[rule] << handler_proc unless @rules[rule].include?(handler_proc)
      DBus::Binding::bus_add_match(@connection, rule)
    end

    # Remove the handler +handler_proc+ for a match rule previously added using
    # Bus#add_signal_receiver.
    def remove_signal_receiver(handler_proc, signal_name=nil, interface=nil, service=nil, path=nil)
      rule = build_signal_rule(signal_name, interface, service, path)
      return nil unless @rules.has_key?(rule)
      @rules[rule].remove(handler_proc)
      DBus::Binding::bus_remove_match(@conn, rule)
    end

    # Generate a match rule for a signal with the specified arguments. If any of
    # the method arguments is +nil+, its processing (below) is skipped.
    #
    # - The rule "type" is set to "signal"
    # - The rule "interface" is set to the value of the +interface+ argument
    # - The rule "sender" is set as follows: If the given +service+ argument does not
    #   start with ":", and is not "org.freedesktop.DBus", the "sender" value
    #   is set to the owner of the supplied +service+ argument (determined with
    #   a +GetServiceOwner+ call), otherwise, the +service+ value is used
    # - The "path" parameter is set to the value of the +path+ argument
    # - The "member" parameter is set to the value of the +signal_name+ argument
    # 
    def build_signal_rule(signal_name, interface, service, path)
      rule = "type='signal'"
      if interface
        rule += ",interface='%s'" % interface
      end
      if service
        if service[0,1] != ":" && service != "org.freedesktop.DBus"
          bus_service = get_service("org.freedesktop.DBus")
          bus_object = bus_service.get_object("/org/freedesktop/DBus",
                                              "org.freedesktop.DBus")
          service = bus_object.GetServiceOwner(service)
        end
        rule += ",sender='%s'" % service
      end
      if path
        rule += ",path='%s'" % path
      end
      if signal_name
        rule += ",member='%s'" % signal_name
      end
      rule
    end

    # Activates the given service on this bus
    def activate_service(service_name, flags=0)
      DBus::Binding::bus_activate_service(@connection, service_name, flags)
    end

    # Returns the UNIX user ID for the given service
    def get_unix_user(service_name)
      DBus::Binding::bus_get_unix_user(@connection, service_name)
    end

    # Returns the base service name of this bus
    def get_base_service
      DBus::Binding::bus_get_base_service(@connection)
    end

    # Sets the base service name of this bus
    def set_base_service(service_name)
      DBus::Binding::bus_set_base_service(@connection, service_name)
    end

    protected
    def signal_dispatcher(connection, message)
      return HANDLER_RESULT_NOT_YET_HANDLED unless message.get_type == MESSAGE_TYPE_SIGNAL
      interface = message.get_interface
      service = message.get_sender
      path = message.get_path
      member = message.get_member
      rule = build_signal_rule(member, interface, service, path)
      if @rules.has_key?(rule)
        args = [interface, member, service, path, message]
        @rules[rule].each {|f| f.call(*args)}
      end
      HANDLER_RESULT_HANDLED
    end
  end

  # The system-wide message bus
  class SystemBus < Bus
    def initialize(glib_mainloop=true)
      super(Bus::TYPE_SYSTEM, glib_mainloop)
    end
  end

  # The session (current login) message bus
  class SessionBus < Bus
    def initialize(glib_mainloop=true)
      super(Bus::TYPE_SESSION, glib_mainloop)
    end
  end

  # The bus that activated this process (if
  # this process was launched by DBus activation)
  class ActivationBus < Bus
    def initialize(glib_mainloop=true)
      super(Bus::TYPE_ACTIVATION, glib_mainloop)
    end
  end

  # Represents a remote object.
  #
  # A RemoteObject is provided by a RemoteService on a particular Bus. RemoteObjects
  # have member functions, and can be called like normal Ruby objects.
  class RemoteObject
    def initialize(service, object_path, interface)
      @service = service
      @object_path = object_path
      @interface = interface
    end

    # Connect the signal +signal_name+ on this remote object to the
    # supplied handler proc +handler_proc+.
    def connect_to_signal(signal_name, handler_proc)
      @service.get_bus.add_signal_receiver(handler_proc,
                                           signal_name,
                                           @interface,
                                           @service.get_service_name,
                                           @object_path)
    end

    # Implements magic remote method calls
    def method_missing(sym, *args)
      name = sym.id2name
      message = DBus::Binding::DBusMessage.new_method_call(@service.get_service_name,
                                                           @object_path,
                                                           @interface,
                                                           name)
      iter = message.get_iter
      args.each{|a| iter.append(a)}
      reply = @service.get_bus.get_connection.send_with_reply_and_block(message, 5000)
      reply_args = reply.to_a
      return nil if reply_args.empty?
      return reply_args[0] if reply_args.length == 1
      return reply_args
    end
  end

  # Base class for exporting your own Services across the Bus
  #
  # Just inherit from Service, providing the name of your service
  # (e.g. org.designfu.SampleService)
  class Service
    def initialize(name, bus=nil)
      @name = name
      @bus = bus
      @bus = Bus.new if @bus.nil?
      DBus::Binding::bus_acquire_service(@bus.get_connection, @name)
    end

    # The name of this service
    def get_name
      @name
    end

    # The bus this service is defined on
    def get_bus
      @bus
    end
  end

  # Base class for objects that support D-BUS invocation messages
  class DBusCallable
    # Create a new DBusCallable instance on the specified +connection+.
    # +dbus_methods+ is an Array containing a list of symbols for the
    # methods which may be invoked remotely.
    def initialize(connection, dbus_methods=[])
      @connection = connection
      @dbus_methods = dbus_methods
    end

    # Process the method invocation message given in +message+. Returns
    # the reply message.
    def dispatch_message(message)
      dispatch(message.get_member, message.to_a, message)
    end

    # Invoke the method +name+, with arguments +args+, and source invocation
    # request message +source_message+, returning the reply message.
    def dispatch(name, args, source_message)
      unless @dbus_methods.include?(name.to_sym)
        return new_error_reply(source_message, "Method '#{name}' not in allowed list")
      end
      unless self.respond_to?(name)
        return new_error_reply(source_message, "No such method '#{name}'")
      end
      ret = nil
      begin
        args = [source_message, *args]
        ret = self.send(name, *args)
      rescue
        return new_error_reply(source_message, $!.to_s)
      end
      reply = DBus::Binding::DBusMessage.new_method_return(source_message)
      iter = reply.get_iter
      iter.append(ret)
      reply
    end

    # Generate a new error reply from source message +message+, and error string
    # +error_message+.
    def new_error_reply(message, error_message)
      error_name = self.class.to_s.gsub(/::/, '.')
      error_name += '.ERROR'
      DBus::Binding::DBusMessage.new_error(message, error_name, error_message)
    end

    # Called when this object is unregistered from the connection
    def on_unregister(connection)
    end

    # Called when a message arrives on the connection for this object
    def on_message(connection, message)
      reply = dispatch_message(message)
      @connection.send(reply)
      HANDLER_RESULT_HANDLED
    end
  end

  # Base class for exporting your own Objects across the Bus
  #
  # Just inherit from Object and provide a list of the symbol names for methods
  # to share across the bus. These will appear as member functions of your
  # Service object.
  class Object < DBusCallable
    def initialize(object_path, service, dbus_methods=[])
      @connection = service.get_bus.get_connection
      super(@connection, dbus_methods)
      @object_path = object_path
      @service = service
      @connection.register_object_path(@object_path, method(:on_unregister), method(:on_message))
    end

    # Emit the signal +signal_name+ for interface +interface+, and include +args+ in the
    # signal message. 
    def emit_signal(interface, signal_name, *args)
      message = DBus::Binding::DBusMessage.new_signal(@object_path, interface, signal_name)
      iter = message.get_iter
      args.each{|a| iter.append(a)}
      @connection.send(message)
    end
  end

  # An object tree allows you to register a handler for a tree of object paths.
  # This means that literal Ruby objects do not need to be created for each object
  # over the bus, but you can have a virtual tree of objects handled by a single
  # Ruby object. There are two ways to handle method calls on virtual objects:
  #
  # 1. Pass a list of dbus_methods in to initialize. This works just like
  #    DBus::Object, except an object_path is passed as the first argument to
  #    each method, denoting which virtual object the call was made on. If all
  #    the objects in the tree support the same methods, this is the best
  #    approach.
  #
  # 2. Override object_method_called. This allows you to define the valid
  #    methods dynamically on an object by object basis. For example, if
  #    providing an object tree that represented a filesystem heirarchy, you'd
  #    only want an ls method on directory objects, not file objects.
  class ObjectTree < DBusCallable
    def initialize(base_path, service, dbus_methods=[])
      @connection = service.get_bus.get_connection
      super(@connection, dbus_methods)
      @base_path = base_path
      @service = service
      @methods = dbus_methods
      @service.get_bus.get_connection.register_fallback(base_path, method(:on_unregister),
                                                        method(:on_message))
    end

    # Create a new absolute +ObjectPath+ for the given relative +path+
    def relative_path_to_object_path(path)
      ObjectPath.new(@base_path + path)
    end

    # Broadcast the signal +signal_name+ for interface +interface+ for
    # the object identified by +relative_path+
    def broadcast_signal(interface, signal_name, relative_path)
      object_path = relative_path_to_object_path(relative_path)
      message = DBus::Binding::DBusMessage.new_signal(object_path, interface, signal_name)
      @connection.send(message)
    end

    def object_method_called
      raise NotImplementedError, "Not implemented"
    end

    def on_message(connection, message)
      target_object_full_path = message.get_path
      n = @base_path.length
      unless @base_path == target_object_full_path[0,n]
        @connection.send(new_error_reply(message, "Invalid target path: #{target_object_full_path}"))
        return HANDLER_RESULT_HANDLED
      end
      target_object_path = target_object_full_path[n..-1]
      target_method = message.get_member
      target_args = message.to_a
      args = [target_object_path, *target_args]
      @connection.send(dispatch(target_method, args, message))
      HANDLER_RESULT_HANDLED
    end
  end

  # A remote service providing objects.
  #
  # A service is typically a process or application that provides
  # remote objects, but can also be the broadcast service that
  # receives signals from all applications on the Bus.
  class RemoteService
    def initialize(bus, service_name)
      @bus = bus
      @service_name = service_name
    end

    # Bus associated with this service
    def get_bus
      @bus
    end

    # The service name
    def get_service_name
      @service_name
    end

    # Creates an object for the specified +object_path+
    def get_object(object_path, interface)
      RemoteObject.new(self, object_path, interface)
    end
  end

  class ObjectPath < String; end

end
