#-----------------------------------------------------
# D-BUS bindings for Ruby
# (C) Copyright 2004 Leon Breedt
# (C) Copyright 2004 Provenco Group Ltd
#
# Licensed under the same terms as the D-BUS library
#-----------------------------------------------------

#-----------------------------------------------------------------
# Native classes
#
#   The classes below provide any additional implementation for 
#   the corresponding classes defined in dbus.so. Things which
#   belong in here are string munging, convenience methods,
#   to_string implementations, etc. 
#
#   Basically, anything which would be too much of a pain to
#   implement in C.
# 
#-----------------------------------------------------------------
module DBus
module Binding

  class DBusConnection
    #--
    # TODO
    #++
    def set_watch_functions(add_func, remove_func, data) # :nodoc:
    end
    def set_timeout_functions(add_func, remove_func, data) # :nodoc:
    end
    def set_wakeup_main_function(add_func, remove_func, data) # :nodoc:
    end
    def set_data(slot, data) # :nodoc:
    end
    def get_data(slot, data) # :nodoc:
    end
  end

  class DBusMessage
    include Enumerable

    # Yields each message argument value that was appended to the message to
    # the given block, in sequence
    def each
      raise DBusError, "block expected" unless block_given?
      iter = get_iter
      begin
        value = iter.value
      rescue
        return nil
      end
      yield value
      have_value = iter.next
      while have_value
        yield iter.value
        have_value = iter.next
      end
      nil
    end

    # Returns the message argument value at the given offset
    def [](offset)
      entries[offset]
    end

    # Appends each item in the given Array to the message
    def append(ary)
      iter = get_append_iter
      ary.each{|value| iter.append(value)}
      nil
    end

    # Appends the given value to the message. If you want to append more than
    # one value, use DBusMessage#append instead, it will be more efficient.
    def <<(value)
      iter = get_append_iter
      iter.append(value)
      nil
    end

    # Returns an array containing the path segments of the message path
    def get_path_decomposed
      get_path.split("/")
    end

    def to_s
      "#<#{self.class.to_s} path=\"#{get_path}\" interface=\"#{get_interface}\" member=\"#{get_member}\" sender=\"#{get_sender}\">"
    end
  end

  class DBusMessageIter
    # Appends the given value to the message this iterator is associated with.
    # Ruby types are automatically converted to the closest matching D-BUS
    # message argument type.
    def append(value)
      case
      when value.nil?
        append_nil
      when value.is_a?(ObjectPath)
        append_object_path(value)
      when value.is_a?(String)
        append_string(value)
      when value.is_a?(Fixnum)
        append_int32(value)
      when value.is_a?(Bignum)
        append_int64(value)
      when value.is_a?(Float)
        append_double(value)
      when value.is_a?(TrueClass)
        append_boolean(true)
      when value.is_a?(FalseClass)
        append_boolean(false)
      when value.is_a?(Hash)
        di = append_dict
        value.each{|k,v| di.append_dict_key(k); di.append(v); }
        true
      when value.is_a?(Array)
        raise ArgumentError, "empty arrays are not supported" if value.empty?
        dtype = ruby_to_dbus_type(value[0])
        ai = append_array(dtype)
        value.each{|v| ai.append(v) if ruby_to_dbus_type(v) == dtype}
        true
      else
        raise ArgumentError, "unsupported argument type #{value.class}"
      end
    end

    # Returns the message argument value at the current position of the
    # iterator. The value is converted from the D-BUS type to the closest
    # matching Ruby type.
    def value
      case get_arg_type
      when TYPE_NIL
        nil
      when TYPE_BYTE
        get_byte
      when TYPE_BOOLEAN
        get_boolean
      when TYPE_INT32
        get_int32
      when TYPE_UINT32
        get_uint32
      when TYPE_INT64
        get_int64
      when TYPE_UINT64
        get_uint64
      when TYPE_DOUBLE
        get_double
      when TYPE_STRING
        get_string
      when TYPE_ARRAY
        get_array
      when TYPE_DICT
        get_dict
      when TYPE_OBJECT_PATH
        ObjectPath.new(get_object_path)
      else
        raise DBusError, "unsupported iter arg type '#{type_as_string(get_arg_type)}'"
      end
    end

    # For compatibility with the other bindings
    alias :get :value

    # Returns the message argument value at the current position of the
    # iterator as a Hash. For this to work, the argument value has to be
    # of the D-BUS type TYPE_DICT.
    def get_dict
      iter = dict_iter
      ret = {}
      loop do
        ret[iter.get_dict_key] = iter.value
        break if !iter.has_next
        iter.next
      end
      ret
    end

    # Returns the message argument value at the current position of the
    # iterator as an Array.  For this to work, the argument value has to be
    # of the D-BUS type TYPE_ARRAY.
    def get_array
      iter = array_iter
      ret = []
      loop do
        ret << iter.value
        break if !iter.has_next
        iter.next
      end
      ret
    end

    private
    def type_as_string(arg_type)
      case arg_type
      when TYPE_INVALID
        'invalid'
      when TYPE_NIL
        'nil'
      when TYPE_BYTE
        'byte'
      when TYPE_BOOLEAN
        'boolean'
      when TYPE_INT32
        'int32'
      when TYPE_UINT32
        'uint32'
      when TYPE_INT64
        'int64'
      when TYPE_UINT64
        'uint64'
      when TYPE_DOUBLE
        'double'
      when TYPE_STRING
        'string'
      when TYPE_CUSTOM
        'custom'
      when TYPE_ARRAY
        'array'
      when TYPE_DICT
        'dict'
      when TYPE_OBJECT_PATH
        'objectpath'
      else
        'unknown'
      end
    end

    def ruby_to_dbus_type(value)
      case
      when value.nil?
        TYPE_NIL
      when value.is_a?(ObjectPath)
        TYPE_OBJECT_PATH
      when value.is_a?(String)
        TYPE_STRING
      when value.is_a?(Fixnum)
        TYPE_INT32
      when value.is_a?(Bignum)
        TYPE_INT64
      when value.is_a?(Float)
        TYPE_DOUBLE
      when value.is_a?(TrueClass)
        TYPE_BOOLEAN
      when value.is_a?(FalseClass)
        TYPE_BOOLEAN
      when value.is_a?(Hash)
        TYPE_DICT
      when value.is_a?(Array)
        TYPE_ARRAY
      else
        raise ArgumentError, "unsupported Ruby type #{value.class}"
      end
    end
  end

end # Binding
end # DBus
