#!/usr/bin/ruby

# Volume Manager example ported from HAL

require 'dbus'
require 'gtk2'

# This just watches the CD-ROM volumes on the system, and prints messages when
# they are added or removed (tray opened/closed). At startup, it adds all the
# existing ones.
#
# This requires udev with D-BUS enabled and HAL to work correctly
#

class HalDevice
  attr :service
  attr :udi
  attr :object

  def initialize(service, udi_or_object, properties={})
    @service = service
    if !udi_or_object.is_a?(String)
      @udi = properties['info.udi']
      @object = udi_or_object
      @properties = properties
    else
      @udi = udi_or_object
      @object = @service.get_object(udi_or_object, 'org.freedesktop.Hal.Device')
      @properties = @object.GetAllProperties()
    end
  end

  def [](name)
    @properties[name]
  end
end

class StorageDevice < HalDevice
  def drive_type
    self['storage.drive_type']
  end

  def device_node
    self['block.device']
  end

  def to_s
    "#<StorageDevice drive_type=\"#{self.drive_type}\" device_node=\"#{self.device_node}\">"
  end
end

class Volume < HalDevice
  attr :storage

  def initialize(service, udi_or_object, properties={}, storage=nil)
    super(service, udi_or_object, properties)
    @storage = storage.nil?? StorageDevice.new(@service, self['block.storage_device']) : storage
  end

  def label
    self['volume.label']
  end

  def mount_point
    self['volume.mount_point']
  end

  def to_s
    "#<Volume mount_point=\"#{self.mount_point}\" storage=#{@storage}>"
  end
end

class CDRomVolume < Volume
  def has_data?
    if self['volume.disc.has_data'] == 1
      return true
    else
      return false
    end
  end

  def has_audio?
    if self['volume.disc.has_audio'] == 1
      return true
    else
      return false
    end
  end

  def to_s
    "#<CDRomVolume label=\"#{self.label}\" mount_point=\"#{self.mount_point}\" storage=#{self.storage}>"
  end
end

class VolumeManager
  def initialize
    @bus = DBus::SystemBus.new
    @hal_service = @bus.get_service('org.freedesktop.Hal')
    @hal_manager = @hal_service.get_object('/org/freedesktop/Hal/Manager',
                                           'org.freedesktop.Hal.Manager')
    connect_gdl_signal(method(:on_gdl_device_added), 'DeviceAdded')
    connect_gdl_signal(method(:on_gdl_device_removed), 'DeviceRemoved')
    @volumes = {}
    device_names = @hal_manager.GetAllDevices()
    device_names.each{|d| add_device(d)}
  end

  def add_device(udi)
    volume_obj = @hal_service.get_object(udi, 'org.freedesktop.Hal.Device')
    volume_prop = volume_obj.GetAllProperties()
    # this causes hald to segfault (!!)
    # return nil unless volume_obj.QueryCapability('volume')
    return nil unless volume_prop['info.category'] == 'volume'
    storage_device = StorageDevice.new(@hal_service, volume_prop['block.storage_device'])
    return nil unless storage_device.drive_type == 'cdrom'
    volume = CDRomVolume.new(@hal_service, volume_obj, volume_prop, storage_device)
    @volumes[volume.udi] = volume
    puts "Add Volume: #{volume}"
  end

  def remove_device(udi)
    return nil unless @volumes.has_key?(udi)
    volume = @volumes.delete(udi)
    puts "Remove Volume: #{volume}"
  end

  def connect_gdl_signal(handler_proc, member)
    @bus.add_signal_receiver(handler_proc,
                             member,                         # member
                             'org.freedesktop.Hal.Manager',  # interface
                             'org.freedesktop.Hal',          # service
                             '/org/freedesktop/Hal/Manager') # path
  end

  def on_gdl_device_added(dbus_if, member, svc, obj_path, message)
    add_device(message.get_iter.value)
  end

  def on_gdl_device_removed(dbus_if, member, svc, obj_path, message)
    remove_device(message.get_iter.value)
  end

  def run
    Gtk.main
  end
end

if __FILE__ == $0
  manager = VolumeManager.new
  manager.run
end
